@@ -6,6 +6,10 @@
 #define _XOPEN_SOURCE 600
 #endif
 
+#ifdef HAVE_LIBFDB5
+#include "cdi_fdb.h"
+#endif
+
 #include <sys/stat.h> // struct stat
 #include <ctype.h>
 
@@ -356,13 +360,15 @@ const char *streamFilename(int streamID)
 static
 long cdiInqTimeSize(int streamID)
 {
-  int tsID = 0, nrecs;
   stream_t *streamptr = stream_to_pointer(streamID);
   long ntsteps = streamptr->ntsteps;
 
-  if ( ntsteps == (long)CDI_UNDEFID )
-    while ( (nrecs = streamInqTimestep(streamID, tsID++)) )
-      ntsteps = streamptr->ntsteps;
+  if (ntsteps == (long)CDI_UNDEFID)
+    {
+      int tsID = 0;
+      while (streamInqTimestep(streamID, tsID++))
+        ntsteps = streamptr->ntsteps;
+    }
 
   return ntsteps;
 }
@@ -371,7 +377,7 @@ static
 int cdiInqContents(stream_t *streamptr)
 {
   int status = 0;
-  int filetype = streamptr->filetype;
+  const int filetype = streamptr->filetype;
 
   switch (cdiBaseFiletype(filetype))
     {
@@ -379,7 +385,7 @@ int cdiInqContents(stream_t *streamptr)
     case CDI_FILETYPE_GRB:
     case CDI_FILETYPE_GRB2:
       {
-        status = grbInqContents(streamptr);
+        status = (streamptr->fdbRetrieve) ? fdbInqContents(streamptr) : grbInqContents(streamptr);
 	break;
       }
 #endif
@@ -441,13 +447,76 @@ int cdiStreamOpenDefaultDelegate(const char *filename, char filemode, int filety
   int fileID;
   char temp[2] = { filemode, 0 };
 
+  const int pathlen = (int)strlen(filename);
+  const bool fdbRetrieve = (pathlen >= 4 && filemode == 'r' && memcmp(filename, "fdb:", 4) == 0);
+  const bool fdbStore    = (pathlen >= 4 && filemode == 'w' && memcmp(filename, "fdb:", 4) == 0);
+
+#ifndef HAVE_LIBFDB5
+  if (fdbRetrieve || fdbStore)
+    {
+      Message("FDB5 support not compiled in!");
+      return CDI_ELIBNAVAIL;
+    }
+#endif
+
+#ifndef HAVE_LIBGRIB_API
+  if (fdbRetrieve || fdbStore)
+    {
+      Message("ecCodes support not compiled in (Needed for FDB5)!");
+      return CDI_ELIBNAVAIL;
+    }
+#endif
+
+  if (fdbRetrieve)
+    {
+      if (filetype == CDI_FILETYPE_GRB || filetype == CDI_FILETYPE_GRB2)
+        {
+#ifdef HAVE_LIBFDB5
+          fdb_initialise();
+          fdb_new_handle((fdb_handle_t **)&(streamptr->fdbHandle));
+          streamptr->fdbRetrieve = true;
+#endif
+          streamptr->filetype = filetype;
+          if (recordBufIsToBeCreated)
+            {
+              streamptr->record = (Record *) Malloc(sizeof(Record));
+              streamptr->record->buffer = NULL;
+            }
+          return 88;
+        }
+      else
+        {
+          Message("Input format must be GRIB to retrieve data to FDB.");
+          return -2;
+        }
+    }
+
+  if (fdbStore)
+    {
+      if (filetype == CDI_FILETYPE_GRB || filetype == CDI_FILETYPE_GRB2)
+        {
+#ifdef HAVE_LIBFDB5
+          fdb_initialise();
+          fdb_new_handle((fdb_handle_t **)&(streamptr->fdbHandle));
+          streamptr->fdbStore = true;
+#endif
+          streamptr->filetype = filetype;
+          return 99;
+        }
+      else
+        {
+          Message("Output format must be GRIB to store data to FDB.");
+          return -2;
+        }
+    }
+
   switch (filetype)
     {
 #if defined(HAVE_LIBGRIB) && (defined(HAVE_LIBCGRIBEX) || defined(HAVE_LIBGRIB_API))
     case CDI_FILETYPE_GRB:
       {
         fileID = gribOpen(filename, temp);
-        if ( fileID < 0 ) return CDI_ESYSTEM;
+        if (fileID < 0) return CDI_ESYSTEM;
         if (recordBufIsToBeCreated)
           {
             streamptr->record = (Record *) Malloc(sizeof(Record));
@@ -464,7 +533,7 @@ int cdiStreamOpenDefaultDelegate(const char *filename, char filemode, int filety
     case CDI_FILETYPE_GRB2:
       {
         fileID = gribOpen(filename, temp);
-        if ( fileID < 0 ) return CDI_ESYSTEM;
+        if (fileID < 0) return CDI_ESYSTEM;
         if (recordBufIsToBeCreated)
           {
             streamptr->record = (Record *) Malloc(sizeof(Record));
@@ -546,10 +615,10 @@ int cdiStreamOpenDefaultDelegate(const char *filename, char filemode, int filety
 
 int streamOpenID(const char *filename, char filemode, int filetype, int resH)
 {
-  if ( CDI_Debug )
-    Message("Open %s mode %c file %s", strfiletype(filetype), filemode, filename?filename:"(NUL)");
+  if (CDI_Debug)
+    Message("Open %s mode %c file %s", strfiletype(filetype), filemode, filename ? filename : "(NUL)");
 
-  if ( ! filename || filetype < 0 ) return CDI_EINVAL;
+  if (! filename || filetype < 0) return CDI_EINVAL;
 
   stream_t *streamptr = stream_new_entry(resH);
   int streamID = CDI_ESYSTEM;
@@ -559,30 +628,30 @@ int streamOpenID(const char *filename, char filemode, int filetype, int resH)
     = (int (*)(const char *, char, int, stream_t *, int))
     namespaceSwitchGet(NSSWITCH_STREAM_OPEN_BACKEND).func;
 
-  int fileID = streamOpenDelegate(filename, filemode, filetype, streamptr, 1);
-  if ( fileID < 0 )
+  const int fileID = streamOpenDelegate(filename, filemode, filetype, streamptr, 1);
+  if (fileID < 0)
     {
       streamID = fileID;
     }
   else
     {
       streamID = streamptr->self;
-      if ( streamID < 0 ) return CDI_ELIMIT;
+      if (streamID < 0) return CDI_ELIMIT;
 
       streamptr->filemode = filemode;
       streamptr->filename = strdupx(filename);
-      streamptr->fileID   = fileID;
+      streamptr->fileID = fileID;
 
-      if ( filemode == 'r' )
+      if (filemode == 'r')
         {
-          int vlistID = vlistCreate();
-          if ( vlistID < 0 ) return CDI_ELIMIT;
+          const int vlistID = vlistCreate();
+          if (vlistID < 0) return CDI_ELIMIT;
 
           cdiVlistMakeInternal(vlistID);
           streamptr->vlistID = vlistID;
-          /* cdiReadByteorder(streamID); */
-          int status = cdiInqContents(streamptr);
-          if ( status < 0 )
+          // cdiReadByteorder(streamID);
+          const int status = cdiInqContents(streamptr);
+          if (status < 0)
             {
               streamID = status;
             }
@@ -595,7 +664,7 @@ int streamOpenID(const char *filename, char filemode, int filetype, int resH)
         }
     }
 
-  if ( streamID < 0 )
+  if (streamID < 0)
     {
       Free(streamptr->record);
       stream_delete_entry(streamptr);
@@ -671,7 +740,7 @@ int streamOpenA(const char *filename, const char *filemode, int filetype)
 #endif
       {
         fileID = gribOpen(filename, filemode);
-        if ( fileID != CDI_UNDEFID ) gribContainersNew(streamptr);
+        if (fileID != CDI_UNDEFID) gribContainersNew(streamptr);
 	break;
       }
 #endif
@@ -715,12 +784,12 @@ int streamOpenA(const char *filename, const char *filemode, int filetype)
 #endif
     default:
       {
-	if ( CDI_Debug ) Message("%s support not compiled in!", strfiletype(filetype));
+	if (CDI_Debug) Message("%s support not compiled in!", strfiletype(filetype));
 	return CDI_ELIBNAVAIL;
       }
     }
 
-  if ( fileID == CDI_UNDEFID )
+  if (fileID == CDI_UNDEFID)
     streamID = CDI_UNDEFID;
   else
     streamptr->fileID = fileID;
@@ -770,14 +839,15 @@ int streamOpenRead(const char *filename)
 {
   cdiInitialize();
 
+  const int pathlen = (int)strlen(filename);
+  const bool fdbRetrieve = (pathlen >= 4 && memcmp(filename, "fdb:", 4) == 0);
+
   int byteorder = 0;
-  int filetype = cdiGetFiletype(filename, &byteorder);
+  const int filetype = fdbRetrieve ? CDI_FILETYPE_GRB2 : cdiGetFiletype(filename, &byteorder);
+  if (filetype < 0) return filetype;
 
-  if ( filetype < 0 ) return filetype;
-
-  int streamID = streamOpen(filename, "r", filetype);
-
-  if ( streamID >= 0 )
+  const int streamID = streamOpen(filename, "r", filetype);
+  if (streamID >= 0)
     {
       stream_t *streamptr = stream_to_pointer(streamID);
       streamptr->byteorder = byteorder;
@@ -792,13 +862,11 @@ int streamOpenAppend(const char *filename)
   cdiInitialize();
 
   int byteorder = 0;
-  int filetype = cdiGetFiletype(filename, &byteorder);
+  const int filetype = cdiGetFiletype(filename, &byteorder);
+  if (filetype < 0) return filetype;
 
-  if ( filetype < 0 ) return filetype;
-
-  int streamID = streamOpenA(filename, "a", filetype);
-
-  if ( streamID >= 0 )
+  const int streamID = streamOpenA(filename, "a", filetype);
+  if (streamID >= 0)
     {
       stream_t *streamptr = stream_to_pointer(streamID);
       streamptr->byteorder = byteorder;
@@ -855,7 +923,7 @@ int streamOpenWrite(const char *filename, int filetype)
 }
 
 static
-void streamDefaultValue ( stream_t * streamptr )
+void streamDefaultValue(stream_t * streamptr)
 {
   streamptr->self              = CDI_UNDEFID;
   streamptr->accesstype        = CDI_UNDEFID;
@@ -917,6 +985,12 @@ void streamDefaultValue ( stream_t * streamptr )
   streamptr->cachedTsID        = -1;
   streamptr->jobs              = NULL;
   streamptr->jobManager        = NULL;
+
+  streamptr->fdbRetrieve       = false;
+  streamptr->fdbStore          = false;
+#ifdef HAVE_LIBFDB5
+  streamptr->fdbHandle         = NULL;
+#endif
 }
 
 static
@@ -925,7 +999,7 @@ stream_t *stream_new_entry(int resH)
   cdiInitialize(); /* ***************** make MT version !!! */
 
   stream_t *streamptr = (stream_t *) Malloc(sizeof(stream_t));
-  streamDefaultValue ( streamptr );
+  streamDefaultValue(streamptr);
 
   if (resH == CDI_UNDEFID)
     streamptr->self = reshPut(streamptr, &streamOps);
@@ -941,9 +1015,19 @@ stream_t *stream_new_entry(int resH)
 
 void cdiStreamCloseDefaultDelegate(stream_t *streamptr, int recordBufIsToBeDeleted)
 {
-  int fileID   = streamptr->fileID;
-  int filetype = streamptr->filetype;
-  if ( fileID == CDI_UNDEFID )
+  if (streamptr->fdbStore || streamptr->fdbRetrieve)
+    {
+#ifdef HAVE_LIBFDB5
+      if (streamptr->fdbHandle) fdb_delete_handle(streamptr->fdbHandle);
+      streamptr->fdbHandle = NULL;
+#endif
+      return;
+    }
+
+  const int fileID = streamptr->fileID;
+  const int filetype = streamptr->filetype;
+
+  if (fileID == CDI_UNDEFID)
     Warning("File %s not open!", streamptr->filename);
   else
     switch (cdiBaseFiletype(filetype))
@@ -993,7 +1077,7 @@ void cdiStreamCloseDefaultDelegate(stream_t *streamptr, int recordBufIsToBeDelet
       case CDI_FILETYPE_NETCDF:
         {
           cdfClose(fileID);
-          if (streamptr->ntsteps == 0)
+          if (streamptr->ntsteps == 0 && streamptr->tsteps != NULL)
             {
               if ( streamptr->tsteps[0].records )
                 {
@@ -1045,8 +1129,7 @@ void streamClose(int streamID)
 {
   stream_t *streamptr = stream_to_pointer(streamID);
 
-  if ( CDI_Debug )
-    Message("streamID = %d filename = %s", streamID, streamptr->filename);
+  if (CDI_Debug) Message("streamID = %d filename = %s", streamID, streamptr->filename);
 
   const int vlistID  = streamptr->vlistID;
 
@@ -1054,11 +1137,11 @@ void streamClose(int streamID)
     = (void (*)(stream_t *, int))
     namespaceSwitchGet(NSSWITCH_STREAM_CLOSE_BACKEND).func;
 
-  if ( streamptr->filetype != -1 ) streamCloseDelegate(streamptr, 1);
+  if (streamptr->filetype != -1) streamCloseDelegate(streamptr, 1);
 
-  if ( streamptr->record )
+  if (streamptr->record)
     {
-      if ( streamptr->record->buffer )
+      if (streamptr->record->buffer)
         Free(streamptr->record->buffer);
 
       Free(streamptr->record);
@@ -1072,36 +1155,35 @@ void streamClose(int streamID)
       sleveltable_t *pslev = streamptr->vars[index].recordTable;
       const unsigned nsub = streamptr->vars[index].subtypeSize >= 0
         ? (unsigned)streamptr->vars[index].subtypeSize : 0U;
-      for (size_t isub=0; isub < nsub; isub++)
-        {
-          deallocate_sleveltable_t(pslev + isub);
-        }
+      for (size_t isub=0; isub < nsub; isub++) deallocate_sleveltable_t(pslev + isub);
       if (pslev) Free(pslev);
     }
   Free(streamptr->vars);
   streamptr->vars = NULL;
 
-  for ( int index = 0; index < streamptr->ntsteps; ++index )
+  for (int index = 0; index < streamptr->ntsteps; ++index)
     {
-      if ( streamptr->tsteps[index].records ) Free(streamptr->tsteps[index].records);
-      if ( streamptr->tsteps[index].recIDs  ) Free(streamptr->tsteps[index].recIDs);
-      taxisDestroyKernel(&streamptr->tsteps[index].taxis);
+      tsteps_t *tstep = &(streamptr->tsteps[index]);
+#ifdef HAVE_LIBFDB5
+      if (tstep->records && tstep->records->fdbItem) free(tstep->records->fdbItem);
+#endif
+      if (tstep->records) Free(tstep->records);
+      if (tstep->recIDs) Free(tstep->recIDs);
+      taxisDestroyKernel(&(tstep->taxis));
     }
 
-  if ( streamptr->tsteps ) Free(streamptr->tsteps);
+  if (streamptr->tsteps) Free(streamptr->tsteps);
 
-  if ( streamptr->basetime.timevar_cache ) Free(streamptr->basetime.timevar_cache);
+  if (streamptr->basetime.timevar_cache) Free(streamptr->basetime.timevar_cache);
 
-  if ( vlistID != -1 )
+  if (vlistID != -1)
     {
-      if ( streamptr->filemode != 'w' && vlistInqTaxis(vlistID) != -1 )
-        taxisDestroy(vlistInqTaxis(vlistID));
-
+      if (streamptr->filemode != 'w' && vlistInqTaxis(vlistID) != -1) taxisDestroy(vlistInqTaxis(vlistID));
       cdiVlistDestroy_(vlistID);
     }
 
   if (streamptr->jobs) free(streamptr->jobs);
-  if (streamptr->jobManager) AsyncWorker_finalize(streamptr->jobManager);
+  if (streamptr->jobManager) AsyncWorker_finalize((AsyncManager *)streamptr->jobManager);
 
   stream_delete_entry(streamptr);
 }
@@ -1286,7 +1368,6 @@ to read the timestamp for this timestep. All calls to read the data refer to thi
 int streamInqTimestep(int streamID, int tsID)
 {
   int nrecs = 0;
-  int taxisID;
   stream_t *streamptr = stream_to_pointer(streamID);
   int vlistID = streamptr->vlistID;
 
@@ -1296,19 +1377,18 @@ int streamInqTimestep(int streamID, int tsID)
       streamptr->curTsID = tsID;
       nrecs = streamptr->tsteps[tsID].nrecs;
       streamptr->tsteps[tsID].curRecID = CDI_UNDEFID;
-      taxisID = vlistInqTaxis(vlistID);
-      if ( taxisID == -1 )
-	Error("Timestep undefined for fileID = %d", streamID);
+      const int taxisID = vlistInqTaxis(vlistID);
+      if (taxisID == -1) Error("Timestep undefined for fileID = %d", streamID);
       ptaxisCopy(taxisPtr(taxisID), &streamptr->tsteps[tsID].taxis);
 
       return nrecs;
     }
 
-  if ( tsID >= streamptr->ntsteps && streamptr->ntsteps != CDI_UNDEFID ) return 0;
+  if (tsID >= streamptr->ntsteps && streamptr->ntsteps != CDI_UNDEFID) return 0;
 
   int filetype = streamptr->filetype;
 
-  if ( CDI_Debug )
+  if (CDI_Debug)
     Message("streamID = %d  tsID = %d  filetype = %d", streamID, tsID, filetype);
 
   switch (cdiBaseFiletype(filetype))
@@ -1317,7 +1397,10 @@ int streamInqTimestep(int streamID, int tsID)
     case CDI_FILETYPE_GRB:
     case CDI_FILETYPE_GRB2:
       {
-        nrecs = grbInqTimestep(streamptr, tsID);
+        if (streamptr->fdbRetrieve)
+          nrecs = fdbInqTimestep(streamptr, tsID);
+        else
+          nrecs = grbInqTimestep(streamptr, tsID);
 	break;
       }
 #endif
@@ -1356,9 +1439,8 @@ int streamInqTimestep(int streamID, int tsID)
       }
     }
 
-  taxisID = vlistInqTaxis(vlistID);
-  if ( taxisID == -1 )
-    Error("Timestep undefined for fileID = %d", streamID);
+  const int taxisID = vlistInqTaxis(vlistID);
+  if (taxisID == -1) Error("Timestep undefined for fileID = %d", streamID);
 
   ptaxisCopy(taxisPtr(taxisID), &streamptr->tsteps[tsID].taxis);
 
@@ -1808,7 +1890,7 @@ streamPack(void * streamptr, void * packBuffer, int packBufferSize,
   uint32_t d = cdiCheckSum(CDI_DATATYPE_INT, streamNint, intBuffer);
   serializePack(&d, 1, CDI_DATATYPE_UINT32, packBuffer, packBufferSize, packBufferPos, context);
 
-  serializePack(&CDI_default_missval, 1, CDI_DATATYPE_FLT64, packBuffer, packBufferSize, packBufferPos, context);
+  serializePack(&CDI_Default_Missval, 1, CDI_DATATYPE_FLT64, packBuffer, packBufferSize, packBufferPos, context);
   serializePack(streamP->filename, intBuffer[2], CDI_DATATYPE_TXT, packBuffer, packBufferSize, packBufferPos, context);
   d = cdiCheckSum(CDI_DATATYPE_TXT, intBuffer[2], streamP->filename);
   serializePack(&d, 1, CDI_DATATYPE_UINT32, packBuffer, packBufferSize, packBufferPos, context);
@@ -1829,7 +1911,7 @@ streamUnpack(char * unpackBuffer, int unpackBufferSize,
   xassert(cdiCheckSum(CDI_DATATYPE_INT, streamNint, intBuffer) == d);
 
   serializeUnpack(unpackBuffer, unpackBufferSize, unpackBufferPos,
-                  &CDI_default_missval, 1, CDI_DATATYPE_FLT64, context);
+                  &CDI_Default_Missval, 1, CDI_DATATYPE_FLT64, context);
   serializeUnpack(unpackBuffer, unpackBufferSize, unpackBufferPos,
                   &filename, intBuffer[2], CDI_DATATYPE_TXT, context);
   serializeUnpack(unpackBuffer, unpackBufferSize, unpackBufferPos,
